import inspect
from collections.abc import Sequence
from typing_extensions import TypeIs

from ..context import DepsT, _utils as _context_utils
from ..messages import (
    AssistantMessage,
    Message,
    SystemMessage,
    UserContent,
    UserMessage,
    user,
)
from ..types import P
from .protocols import (
    AsyncContextMessageTemplate,
    AsyncMessageTemplate,
    ContextMessageTemplate,
    MessageTemplate,
)


def is_messages(
    messages_or_content: Sequence[Message] | UserContent,
) -> TypeIs[Sequence[Message]]:
    if isinstance(messages_or_content, list):
        if not messages_or_content:
            raise ValueError("Empty array may not be used as message content")
        return isinstance(
            messages_or_content[0], SystemMessage | UserMessage | AssistantMessage
        )
    return False


def promote_to_messages(result: Sequence[Message] | UserContent) -> Sequence[Message]:
    """Promote a prompt result to a list of messages.

    If the result is already a list of Messages, returns it as-is.
    If the result is UserContent, wraps it in a single user message.
    """
    if is_messages(result):
        return result
    return [user(result)]


def is_context_promptable(
    fn: ContextMessageTemplate[P, DepsT]
    | AsyncContextMessageTemplate[P, DepsT]
    | MessageTemplate[P]
    | AsyncMessageTemplate[P],
) -> TypeIs[ContextMessageTemplate[P, DepsT] | AsyncContextMessageTemplate[P, DepsT]]:
    """Type guard to check if a function is a context promptable function."""
    return _context_utils.first_param_is_context(fn)


def is_async_promptable(
    fn: ContextMessageTemplate[P, DepsT]
    | AsyncContextMessageTemplate[P, DepsT]
    | MessageTemplate[P]
    | AsyncMessageTemplate[P],
) -> TypeIs[AsyncMessageTemplate[P] | AsyncContextMessageTemplate[P, DepsT]]:
    """Type guard to check if a function is an async promptable function."""
    return inspect.iscoroutinefunction(fn)



"""Beta Anthropic message encoding and request preparation."""

from collections.abc import Sequence
from typing import Any, TypedDict, cast
from typing_extensions import Required

from anthropic import Omit
from anthropic.types.anthropic_beta_param import AnthropicBetaParam
from anthropic.types.beta import (
    BetaContentBlockParam,
    BetaMessageParam,
    BetaTextBlockParam,
    BetaThinkingConfigParam,
    BetaToolChoiceParam,
    BetaToolParam,
)
from pydantic import BaseModel

from ....content import ContentPart
from ....exceptions import FormattingModeNotSupportedError
from ....formatting import (
    Format,
    FormattableT,
    resolve_format,
)
from ....messages import AssistantMessage, Message, UserMessage
from ....tools import AnyToolSchema, BaseToolkit
from ...base import Params, _utils as _base_utils
from ..model_id import model_name
from ..model_info import MODELS_WITHOUT_STRICT_STRUCTURED_OUTPUTS
from .encode import (
    DEFAULT_MAX_TOKENS,
    FORMAT_TOOL_NAME,
    convert_tool_to_tool_param,
    encode_content,
    process_params,
)

DEFAULT_FORMAT_MODE = "strict"


class BetaParseKwargs(TypedDict, total=False):
    """Kwargs for Anthropic beta.messages.parse method."""

    model: Required[str]
    max_tokens: Required[int]
    messages: Sequence[BetaMessageParam]
    system: Sequence[BetaTextBlockParam] | Omit
    tools: Sequence[BetaToolParam] | Omit
    tool_choice: BetaToolChoiceParam | Omit
    temperature: float | Omit
    top_p: float | Omit
    top_k: int | Omit
    stop_sequences: list[str] | Omit
    thinking: BetaThinkingConfigParam | Omit
    betas: list[AnthropicBetaParam]
    output_format: type[BaseModel]


def _beta_encode_content(
    content: Sequence[ContentPart],
    encode_thoughts_as_text: bool,
    add_cache_control: bool = False,
) -> str | Sequence[BetaContentBlockParam]:
    """Convert mirascope content to Beta Anthropic content format."""
    result = encode_content(content, encode_thoughts_as_text, add_cache_control)
    if isinstance(result, str):
        return result
    return cast(Sequence[BetaContentBlockParam], result)


def _beta_encode_message(
    message: UserMessage | AssistantMessage,
    model_id: str,
    encode_thoughts_as_text: bool,
    add_cache_control: bool = False,
) -> BetaMessageParam:
    """Convert user or assistant Message to Beta MessageParam format.

    Args:
        message: The message to encode
        model_id: The Anthropic model ID
        encode_thoughts_as_text: Whether to encode thought blocks as text
        add_cache_control: Whether to add cache_control to the last content block
    """
    if (
        message.role == "assistant"
        and message.provider_id == "anthropic"
        and message.model_id == model_id
        and message.raw_message
        and not encode_thoughts_as_text
        and not add_cache_control
    ):
        raw = cast(dict[str, Any], message.raw_message)
        return BetaMessageParam(
            role=raw["role"],
            content=raw["content"],
        )

    content = _beta_encode_content(
        message.content, encode_thoughts_as_text, add_cache_control
    )

    return BetaMessageParam(
        role=message.role,
        content=content,
    )


def _beta_encode_messages(
    messages: Sequence[UserMessage | AssistantMessage],
    model_id: str,
    encode_thoughts_as_text: bool,
) -> Sequence[BetaMessageParam]:
    """Encode messages and add cache control for multi-turn conversations.

    If the conversation contains assistant messages (indicating multi-turn),
    adds cache_control to the last content block of the last message.
    """
    # Detect multi-turn conversations by checking for assistant messages
    has_assistant_message = any(msg.role == "assistant" for msg in messages)

    # Encode messages, adding cache_control to the last message if multi-turn
    encoded_messages: list[BetaMessageParam] = []
    for i, message in enumerate(messages):
        is_last = i == len(messages) - 1
        add_cache = has_assistant_message and is_last
        encoded_messages.append(
            _beta_encode_message(message, model_id, encode_thoughts_as_text, add_cache)
        )
    return encoded_messages


def _beta_convert_tool_to_tool_param(tool: AnyToolSchema) -> BetaToolParam:
    """Convert a single Mirascope tool to Beta Anthropic tool format."""
    return cast(BetaToolParam, convert_tool_to_tool_param(tool))


def beta_encode_request(
    *,
    model_id: str,
    messages: Sequence[Message],
    tools: Sequence[AnyToolSchema] | BaseToolkit[AnyToolSchema] | None,
    format: type[FormattableT] | Format[FormattableT] | None,
    params: Params,
) -> tuple[Sequence[Message], Format[FormattableT] | None, BetaParseKwargs]:
    """Prepares a request for the Anthropic beta.messages.parse method."""

    processed = process_params(params, DEFAULT_MAX_TOKENS)
    encode_thoughts_as_text = processed.pop("encode_thoughts_as_text", False)
    max_tokens = processed.pop("max_tokens", DEFAULT_MAX_TOKENS)

    kwargs: BetaParseKwargs = BetaParseKwargs(
        {
            "model": model_name(model_id),
            "max_tokens": max_tokens,
            "betas": ["structured-outputs-2025-11-13"],
            **processed,
        }
    )

    tools = tools.tools if isinstance(tools, BaseToolkit) else tools or []
    anthropic_tools = [_beta_convert_tool_to_tool_param(tool) for tool in tools]
    format = resolve_format(format, default_mode=DEFAULT_FORMAT_MODE)

    if format is not None:
        if format.mode == "strict":
            if model_name(model_id) in MODELS_WITHOUT_STRICT_STRUCTURED_OUTPUTS:
                raise FormattingModeNotSupportedError(
                    formatting_mode=format.mode,
                    provider_id="anthropic",
                    model_id=model_id,
                )
            else:
                kwargs["output_format"] = cast(type[BaseModel], format.formattable)

        if format.mode == "tool":
            format_tool_schema = format.create_tool_schema()
            anthropic_tools.append(_beta_convert_tool_to_tool_param(format_tool_schema))
            if tools:
                kwargs["tool_choice"] = {"type": "any"}
            else:
                kwargs["tool_choice"] = {
                    "type": "tool",
                    "name": FORMAT_TOOL_NAME,
                    "disable_parallel_tool_use": True,
                }

        if format.formatting_instructions:
            messages = _base_utils.add_system_instructions(
                messages, format.formatting_instructions
            )

    if anthropic_tools:
        # Add cache control to the last tool for prompt caching
        last_tool = anthropic_tools[-1]
        last_tool["cache_control"] = {"type": "ephemeral"}
        kwargs["tools"] = anthropic_tools

    system_message_content, remaining_messages = _base_utils.extract_system_message(
        messages
    )

    kwargs["messages"] = _beta_encode_messages(
        remaining_messages, model_id, encode_thoughts_as_text
    )

    if system_message_content:
        kwargs["system"] = [
            BetaTextBlockParam(
                type="text",
                text=system_message_content,
                cache_control={"type": "ephemeral"},
            )
        ]

    return messages, format, kwargs



"""Tests for MirascopeProvider and utilities."""

import os
from unittest.mock import AsyncMock, Mock, patch

import pytest

from mirascope.llm.providers.mirascope import _utils
from mirascope.llm.providers.mirascope.provider import MirascopeProvider


class TestMirascopeUtils:
    """Tests for Mirascope utility functions."""

    def test_extract_provider_prefix_valid(self) -> None:
        """Test extracting provider prefix from valid model IDs."""
        assert _utils.extract_provider_prefix("openai/gpt-4") == "openai"
        assert _utils.extract_provider_prefix("anthropic/claude-3") == "anthropic"
        assert _utils.extract_provider_prefix("google/gemini-pro") == "google"
        assert (
            _utils.extract_provider_prefix("openai/gpt-4-with-extra/stuff") == "openai"
        )

    def test_extract_provider_prefix_invalid(self) -> None:
        """Test extracting provider prefix from invalid model IDs."""
        assert _utils.extract_provider_prefix("gpt-4") is None
        assert _utils.extract_provider_prefix("") is None
        assert _utils.extract_provider_prefix("no-slash") is None

    def test_get_default_router_base_url_default(self) -> None:
        """Test getting default router base URL."""
        original_url = os.environ.pop("MIRASCOPE_ROUTER_BASE_URL", None)
        try:
            url = _utils.get_default_router_base_url()
            assert url == "https://mirascope.com/router/v0"
        finally:
            if original_url is not None:
                os.environ["MIRASCOPE_ROUTER_BASE_URL"] = original_url

    def test_get_default_router_base_url_from_env(self) -> None:
        """Test getting router base URL from environment variable."""
        original_url = os.environ.get("MIRASCOPE_ROUTER_BASE_URL")
        os.environ["MIRASCOPE_ROUTER_BASE_URL"] = "http://localhost:3000/router/v0"
        try:
            url = _utils.get_default_router_base_url()
            assert url == "http://localhost:3000/router/v0"
        finally:
            if original_url is not None:
                os.environ["MIRASCOPE_ROUTER_BASE_URL"] = original_url
            else:
                os.environ.pop("MIRASCOPE_ROUTER_BASE_URL", None)

    def test_create_underlying_provider_openai(self) -> None:
        """Test creating OpenAI provider."""
        provider = _utils.create_underlying_provider(
            provider_prefix="openai",
            api_key="test-key",
            router_base_url="http://localhost:3000/router/v0",
        )
        assert provider.id == "openai"
        assert "localhost:3000/router/v0/openai" in str(provider.client.base_url)

    def test_create_underlying_provider_anthropic(self) -> None:
        """Test creating Anthropic provider."""
        provider = _utils.create_underlying_provider(
            provider_prefix="anthropic",
            api_key="test-key",
            router_base_url="http://localhost:3000/router/v0",
        )
        assert provider.id == "anthropic"
        assert "localhost:3000/router/v0/anthropic" in str(provider.client.base_url)

    def test_create_underlying_provider_google(self) -> None:
        """Test creating Google provider."""
        provider = _utils.create_underlying_provider(
            provider_prefix="google",
            api_key="test-key",
            router_base_url="http://localhost:3000/router/v0",
        )
        assert provider.id == "google"

    def test_create_underlying_provider_unsupported(self) -> None:
        """Test creating provider with unsupported prefix."""
        with pytest.raises(ValueError) as exc_info:
            _utils.create_underlying_provider(
                provider_prefix="unknown",
                api_key="test-key",
                router_base_url="http://localhost:3000/router/v0",
            )
        assert "Unsupported provider: unknown" in str(exc_info.value)
        assert "anthropic, google, openai" in str(exc_info.value)

    def test_create_underlying_provider_caching(self) -> None:
        """Test that provider creation is cached."""
        provider1 = _utils.create_underlying_provider(
            provider_prefix="openai",
            api_key="test-key",
            router_base_url="http://localhost:3000/router/v0",
        )
        provider2 = _utils.create_underlying_provider(
            provider_prefix="openai",
            api_key="test-key",
            router_base_url="http://localhost:3000/router/v0",
        )
        # Should be the exact same instance due to caching
        assert provider1 is provider2

        # Different parameters should create different instances
        provider3 = _utils.create_underlying_provider(
            provider_prefix="openai",
            api_key="different-key",
            router_base_url="http://localhost:3000/router/v0",
        )
        assert provider1 is not provider3


class TestMirascopeProvider:
    """Tests for MirascopeProvider."""

    def test_mirascope_provider_initialization_with_api_key(self) -> None:
        """Test MirascopeProvider initialization with api_key."""
        provider = MirascopeProvider(api_key="test-api-key")
        assert provider.id == "mirascope"
        assert provider.default_scope == ["anthropic/", "google/", "openai/"]
        assert provider.api_key == "test-api-key"
        assert provider.router_base_url == "https://mirascope.com/router/v0"

    def test_mirascope_provider_missing_api_key(self) -> None:
        """Test MirascopeProvider raises error when API key is missing."""
        original_key = os.environ.pop("MIRASCOPE_API_KEY", None)
        try:
            with pytest.raises(ValueError) as exc_info:
                MirascopeProvider()
            assert "Mirascope API key not found" in str(exc_info.value)
            assert "MIRASCOPE_API_KEY" in str(exc_info.value)
        finally:
            if original_key is not None:
                os.environ["MIRASCOPE_API_KEY"] = original_key

    def test_mirascope_provider_uses_env_var_api_key(self) -> None:
        """Test MirascopeProvider uses MIRASCOPE_API_KEY from environment."""
        original_key = os.environ.get("MIRASCOPE_API_KEY")
        os.environ["MIRASCOPE_API_KEY"] = "env-test-key"
        try:
            provider = MirascopeProvider()
            assert provider.api_key == "env-test-key"
        finally:
            if original_key is not None:
                os.environ["MIRASCOPE_API_KEY"] = original_key
            else:
                os.environ.pop("MIRASCOPE_API_KEY", None)

    def test_mirascope_provider_custom_base_url(self) -> None:
        """Test MirascopeProvider with custom base_url."""
        provider = MirascopeProvider(
            api_key="test-key", base_url="http://localhost:3000/router/v0"
        )
        assert provider.router_base_url == "http://localhost:3000/router/v0"

    def test_mirascope_provider_uses_env_var_base_url(self) -> None:
        """Test MirascopeProvider uses MIRASCOPE_ROUTER_BASE_URL from environment."""
        original_url = os.environ.get("MIRASCOPE_ROUTER_BASE_URL")
        os.environ["MIRASCOPE_ROUTER_BASE_URL"] = "http://custom:8080/router/v0"
        try:
            provider = MirascopeProvider(api_key="test-key")
            assert provider.router_base_url == "http://custom:8080/router/v0"
        finally:
            if original_url is not None:
                os.environ["MIRASCOPE_ROUTER_BASE_URL"] = original_url
            else:
                os.environ.pop("MIRASCOPE_ROUTER_BASE_URL", None)

    def test_get_underlying_provider_invalid_format(self) -> None:
        """Test _get_underlying_provider with invalid model ID format."""
        provider = MirascopeProvider(api_key="test-key")

        with pytest.raises(ValueError) as exc_info:
            provider._get_underlying_provider("gpt-4")  # pyright: ignore[reportPrivateUsage]
        assert "Invalid model ID format: gpt-4" in str(exc_info.value)
        assert "'provider/model-name'" in str(exc_info.value)

    def test_get_underlying_provider_valid_openai(self) -> None:
        """Test _get_underlying_provider with valid OpenAI model ID."""
        provider = MirascopeProvider(api_key="test-key")
        underlying = provider._get_underlying_provider("openai/gpt-4")  # pyright: ignore[reportPrivateUsage]
        assert underlying.id == "openai"

    def test_get_underlying_provider_valid_anthropic(self) -> None:
        """Test _get_underlying_provider with valid Anthropic model ID."""
        provider = MirascopeProvider(api_key="test-key")
        underlying = provider._get_underlying_provider("anthropic/claude-3")  # pyright: ignore[reportPrivateUsage]
        assert underlying.id == "anthropic"

    def test_get_underlying_provider_valid_google(self) -> None:
        """Test _get_underlying_provider with valid Google model ID."""
        provider = MirascopeProvider(api_key="test-key")
        underlying = provider._get_underlying_provider("google/gemini-pro")  # pyright: ignore[reportPrivateUsage]
        assert underlying.id == "google"

    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    def test_call_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _call delegates to underlying provider."""
        mock_underlying = Mock()
        mock_underlying.call.return_value = Mock()
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        provider._call(model_id="openai/gpt-4", messages=[])  # pyright: ignore[reportPrivateUsage]

        mock_underlying.call.assert_called_once()

    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    def test_context_call_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _context_call delegates to underlying provider."""
        from mirascope.llm.context import Context

        mock_underlying = Mock()
        mock_underlying.context_call.return_value = Mock()
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        ctx = Context(deps={})
        provider._context_call(ctx=ctx, model_id="openai/gpt-4", messages=[])  # pyright: ignore[reportPrivateUsage]

        mock_underlying.context_call.assert_called_once()

    @pytest.mark.asyncio
    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    async def test_call_async_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _call_async delegates to underlying provider."""
        mock_underlying = Mock()
        mock_underlying.call_async = AsyncMock(return_value=Mock())
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        await provider._call_async(model_id="openai/gpt-4", messages=[])  # pyright: ignore[reportPrivateUsage]

        mock_underlying.call_async.assert_called_once()

    @pytest.mark.asyncio
    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    async def test_context_call_async_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _context_call_async delegates to underlying provider."""
        from mirascope.llm.context import Context

        mock_underlying = Mock()
        mock_underlying.context_call_async = AsyncMock(return_value=Mock())
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        ctx = Context(deps={})
        await provider._context_call_async(  # pyright: ignore[reportPrivateUsage]
            ctx=ctx, model_id="openai/gpt-4", messages=[]
        )

        mock_underlying.context_call_async.assert_called_once()

    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    def test_stream_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _stream delegates to underlying provider."""
        mock_underlying = Mock()
        mock_underlying.stream.return_value = Mock()
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        provider._stream(model_id="openai/gpt-4", messages=[])  # pyright: ignore[reportPrivateUsage]

        mock_underlying.stream.assert_called_once()

    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    def test_context_stream_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _context_stream delegates to underlying provider."""
        from mirascope.llm.context import Context

        mock_underlying = Mock()
        mock_underlying.context_stream.return_value = Mock()
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        ctx = Context(deps={})
        provider._context_stream(ctx=ctx, model_id="openai/gpt-4", messages=[])  # pyright: ignore[reportPrivateUsage]

        mock_underlying.context_stream.assert_called_once()

    @pytest.mark.asyncio
    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    async def test_stream_async_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _stream_async delegates to underlying provider."""
        mock_underlying = Mock()
        mock_underlying.stream_async = AsyncMock(return_value=Mock())
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        await provider._stream_async(model_id="openai/gpt-4", messages=[])  # pyright: ignore[reportPrivateUsage]

        mock_underlying.stream_async.assert_called_once()

    @pytest.mark.asyncio
    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    async def test_context_stream_async_delegates_to_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that _context_stream_async delegates to underlying provider."""
        from mirascope.llm.context import Context

        mock_underlying = Mock()
        mock_underlying.context_stream_async = AsyncMock(return_value=Mock())
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")
        ctx = Context(deps={})
        await provider._context_stream_async(  # pyright: ignore[reportPrivateUsage]
            ctx=ctx, model_id="openai/gpt-4", messages=[]
        )

        mock_underlying.context_stream_async.assert_called_once()


class TestMirascopeProviderErrorHandling:
    """Tests for MirascopeProvider error handling."""

    def test_error_map_exists(self) -> None:
        """Test that error_map class variable is defined."""
        assert hasattr(MirascopeProvider, "error_map")
        assert isinstance(MirascopeProvider.error_map, dict)
        assert len(MirascopeProvider.error_map) == 0

    def test_get_error_status_returns_none(self) -> None:
        """Test that get_error_status returns None."""
        provider = MirascopeProvider(api_key="test-key")
        result = provider.get_error_status(Exception("test"))
        assert result is None

    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    def test_error_propagation_from_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that errors from underlying providers propagate correctly."""
        from mirascope.llm.exceptions import RateLimitError

        # Mock underlying provider to raise a Mirascope exception
        mock_underlying = Mock()
        mock_underlying.call.side_effect = RateLimitError("Rate limit exceeded")
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")

        # Should propagate the RateLimitError from underlying provider
        with pytest.raises(RateLimitError) as exc_info:
            provider.call(model_id="openai/gpt-4", messages=[])

        assert "Rate limit exceeded" in str(exc_info.value)

    @pytest.mark.asyncio
    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    async def test_async_error_propagation_from_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that errors from underlying providers propagate correctly in async calls."""
        from mirascope.llm.exceptions import AuthenticationError

        # Mock underlying provider to raise a Mirascope exception
        mock_underlying = Mock()
        mock_underlying.call_async = AsyncMock(
            side_effect=AuthenticationError("Invalid API key")
        )
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")

        # Should propagate the AuthenticationError from underlying provider
        with pytest.raises(AuthenticationError) as exc_info:
            await provider.call_async(model_id="anthropic/claude-3", messages=[])

        assert "Invalid API key" in str(exc_info.value)

    @patch("mirascope.llm.providers.mirascope._utils.create_underlying_provider")
    def test_stream_error_propagation_from_underlying_provider(
        self, mock_create_provider: Mock
    ) -> None:
        """Test that errors from underlying providers propagate correctly in streams."""
        from mirascope.llm.exceptions import ServerError

        # Mock underlying provider to raise a Mirascope exception
        mock_underlying = Mock()
        mock_underlying.stream.side_effect = ServerError("Internal server error")
        mock_create_provider.return_value = mock_underlying

        provider = MirascopeProvider(api_key="test-key")

        # Should propagate the ServerError from underlying provider
        with pytest.raises(ServerError) as exc_info:
            provider.stream(model_id="google/gemini-pro", messages=[])

        assert "Internal server error" in str(exc_info.value)



"""E2E verification of ClickHouse sync via Search API.

Complete end-to-end test:
1. Send spans via OTLP exporter (same as local_versioning_example.py)
2. Wait for ClickHouse sync (polling Search API)
3. Verify data via Search API endpoints

Prerequisites:
1. Start Docker: `cd cloud && docker compose -f docker/compose.yml up -d`
2. Start cloud server: `bun run cloud:dev`
3. Start sync worker: `cd cloud && bun run tsx workers/clickhouseSyncLocal.ts`

Usage:
    MIRASCOPE_API_KEY=mk_xxx uv run python examples/misc/local_clickhouse_verify.py
"""

from __future__ import annotations

import os
import time
from collections.abc import Iterable
from datetime import datetime, timedelta, timezone
from typing import Any

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from mirascope import ops
from mirascope.api._generated.traces import TracesSearchResponse
from mirascope.api.client import Mirascope, create_export_client
from mirascope.ops._internal.exporters import MirascopeOTLPExporter

BASE_URL = os.getenv("MIRASCOPE_BASE_URL", "http://localhost:3000/api/v0")
API_KEY = os.getenv("MIRASCOPE_API_KEY")


def send_spans(provider: TracerProvider) -> tuple[str, set[str]]:
    """Send test spans via OTLP exporter. Returns span name for verification."""

    @ops.trace
    def e2e_clickhouse_test(message: str) -> str:
        """E2E test trace for ClickHouse verification."""
        return f"processed: {message}"

    @ops.trace
    def e2e_nested_operation(data: str) -> str:
        """Nested operation to create multi-span trace."""
        return f"nested: {data}"

    # Create a trace with nested spans
    trace = e2e_clickhouse_test.wrapped("ClickHouse E2E test message")
    print(f"    Trace result: {trace.result}")

    nested = e2e_nested_operation.wrapped("nested data")
    print(f"    Nested result: {nested.result}")

    provider.force_flush()
    print("    Spans flushed!")

    return "e2e_clickhouse_test", {"e2e_clickhouse_test", "e2e_nested_operation"}


def wait_for_sync(
    client: Mirascope, span_name: str, timeout_sec: int = 30
) -> TracesSearchResponse | None:
    """Poll Search API until spans appear or timeout."""
    now = datetime.now(timezone.utc)
    start_time = (now - timedelta(minutes=5)).isoformat()
    end_time = (now + timedelta(minutes=5)).isoformat()

    for i in range(timeout_sec):
        result = client.traces.search(
            start_time=start_time,
            end_time=end_time,
            query=span_name,
            limit=10,
            sort_by="start_time",
            sort_order="desc",
        )
        if result.total and result.total > 0:
            return result
        print(f"    Waiting for sync... ({i + 1}/{timeout_sec}s)")
        time.sleep(1)
    return None


def normalize_datetime(value: str | datetime) -> datetime:
    if isinstance(value, datetime):
        if value.tzinfo is None:
            return value.replace(tzinfo=timezone.utc)
        return value
    parsed = datetime.fromisoformat(value.replace("Z", "+00:00"))
    if parsed.tzinfo is None:
        return parsed.replace(tzinfo=timezone.utc)
    return parsed


def assert_span_fields(span: Any, require_end_time: bool) -> None:
    assert span.name, "span.name is required"
    assert span.trace_id, "span.trace_id is required"
    assert span.span_id, "span.span_id is required"
    assert span.start_time, "span.start_time is required"
    if require_end_time:
        assert span.end_time, "span.end_time is required"
    assert span.duration_ms is not None, "span.duration_ms is required"
    assert span.duration_ms >= 0, "span.duration_ms must be non-negative"
    start_time = normalize_datetime(span.start_time)
    if require_end_time:
        end_time = normalize_datetime(span.end_time)
        assert end_time >= start_time, "span.end_time must be >= span.start_time"


def assert_recent_span(span: Any, now: datetime, window_minutes: int = 10) -> None:
    start_time = normalize_datetime(span.start_time)
    delta = abs((now - start_time).total_seconds())
    assert delta <= window_minutes * 60, (
        f"span.start_time too old/new: {span.start_time}"
    )


def assert_trace_contains_names(spans: Iterable[Any], expected: set[str]) -> None:
    names = {span.name for span in spans if span.name}
    missing = expected - names
    assert not missing, f"missing expected span names: {sorted(missing)}"


def verify_search_api(
    client: Mirascope,
    search_result: TracesSearchResponse,
    expected_names: set[str],
    span_name: str,
) -> bool:
    """Verify Search API endpoints with the synced data."""
    spans = search_result.spans or []
    total = search_result.total or 0
    print(f"    Found {total} spans (showing {len(spans)})")

    for span in spans[:3]:
        assert_span_fields(span, require_end_time=False)
        dur = span.duration_ms
        dur_str = f"{dur}ms" if dur is not None else "NULL"
        model = span.model or "N/A"
        name = span.name[:30] if span.name else "N/A"
        print(f"      - {name:<30} | {model:<15} | {dur_str}")

    if not spans:
        return False

    trace_id = spans[0].trace_id
    trace_result = client.traces.gettracedetail(trace_id)

    if not trace_result:
        print("    Failed to get trace detail")
        return False

    trace_spans = trace_result.spans or []
    print(f"    Trace {trace_id[:16]}... has {len(trace_spans)} spans")
    assert trace_spans, "trace detail has no spans"
    for span in trace_spans:
        assert_span_fields(span, require_end_time=True)
    assert_trace_contains_names(trace_spans, {span_name})
    now = datetime.now(timezone.utc)
    for span in trace_spans:
        assert_recent_span(span, now)

    start_time = (now - timedelta(minutes=10)).isoformat()
    end_time = (now + timedelta(minutes=5)).isoformat()

    analytics_result = client.traces.getanalyticssummary(
        start_time=start_time,
        end_time=end_time,
    )

    if not analytics_result:
        print("    Failed to get analytics")
        return False

    print(
        f"    Analytics: {analytics_result.total_spans or 0} spans, "
        f"avg {analytics_result.avg_duration_ms}ms"
    )
    assert analytics_result.total_spans is not None, "analytics.total_spans missing"
    assert analytics_result.total_spans >= len(trace_spans), (
        "analytics.total_spans is less than trace spans"
    )
    assert analytics_result.avg_duration_ms is not None, (
        "analytics.avg_duration_ms missing"
    )
    assert analytics_result.avg_duration_ms >= 0, (
        "analytics.avg_duration_ms must be non-negative"
    )

    return True


def main():
    if not API_KEY:
        print("ERROR: MIRASCOPE_API_KEY environment variable is required")
        print("Create an API key in the Mirascope Cloud UI: http://localhost:3000")
        exit(1)

    print("=" * 70)
    print("ClickHouse E2E Verification")
    print("=" * 70)
    print(f"API Base URL: {BASE_URL}")
    print(f"API Key: {API_KEY[:15]}...")

    # Create Mirascope client for Search API
    api_client = Mirascope(base_url=BASE_URL, api_key=API_KEY)

    # 1. Setup OTLP exporter
    print("\n[1/4] Setting up OTLP exporter...")
    export_client = create_export_client(base_url=BASE_URL, api_key=API_KEY)
    provider = TracerProvider()
    exporter = MirascopeOTLPExporter(client=export_client)
    provider.add_span_processor(BatchSpanProcessor(exporter))
    ops.configure(tracer_provider=provider)
    ops.instrument_llm()
    print("    OTLP exporter configured")

    # 2. Send spans
    print("\n[2/4] Sending test spans...")
    span_name, expected_names = send_spans(provider)

    # 3. Wait for ClickHouse sync
    print("\n[3/4] Waiting for ClickHouse sync...")
    search_result = wait_for_sync(api_client, span_name)

    if not search_result:
        print("    TIMEOUT: Spans not found in ClickHouse")
        print("    Make sure clickhouseSyncLocal worker is running:")
        print("      cd cloud && bun run tsx workers/clickhouseSyncLocal.ts")
        provider.shutdown()
        exit(1)

    print("    Sync complete!")

    # 4. Verify Search API
    print("\n[4/4] Verifying Search API...")
    success = verify_search_api(api_client, search_result, expected_names, span_name)

    provider.shutdown()

    print("\n" + "=" * 70)
    if success:
        print("E2E Verification PASSED!")
    else:
        print("E2E Verification FAILED!")
        exit(1)
    print("=" * 70)


if __name__ == "__main__":
    main()



"""OpenAI completions message encoding and request preparation."""

from collections.abc import Sequence
from functools import lru_cache
from typing import TypedDict, cast

from openai import Omit
from openai.types import chat as openai_types, shared_params as shared_openai_types
from openai.types.shared_params.response_format_json_schema import JSONSchema

from .....exceptions import (
    FeatureNotSupportedError,
    FormattingModeNotSupportedError,
)
from .....formatting import (
    Format,
    FormattableT,
    resolve_format,
)
from .....messages import AssistantMessage, Message, UserMessage
from .....tools import FORMAT_TOOL_NAME, AnyToolSchema, BaseToolkit
from ....base import Params, _utils as _base_utils
from ...model_id import OpenAIModelId, model_name
from ...model_info import (
    MODELS_WITHOUT_AUDIO_SUPPORT,
    MODELS_WITHOUT_JSON_OBJECT_SUPPORT,
    MODELS_WITHOUT_JSON_SCHEMA_SUPPORT,
)


class ChatCompletionCreateKwargs(TypedDict, total=False):
    """Kwargs for OpenAI ChatCompletion.create method."""

    model: str
    messages: Sequence[openai_types.ChatCompletionMessageParam]
    tools: Sequence[openai_types.ChatCompletionToolParam] | Omit
    response_format: (
        shared_openai_types.ResponseFormatJSONObject
        | shared_openai_types.ResponseFormatJSONSchema
        | Omit
    )
    tool_choice: openai_types.ChatCompletionToolChoiceOptionParam | Omit
    parallel_tool_calls: bool | Omit
    temperature: float | Omit
    max_tokens: int | Omit
    top_p: float | Omit
    seed: int | Omit
    stop: str | list[str] | Omit
    reasoning_effort: shared_openai_types.ReasoningEffort | Omit


def _encode_user_message(
    message: UserMessage,
    model_id: OpenAIModelId,
) -> list[openai_types.ChatCompletionMessageParam]:
    """Convert Mirascope `UserMessage` to a list of OpenAI `ChatCompletionMessageParam`.

    Multiple text content parts are combined into a single user message.
    Tool outputs become separate tool messages.
    """
    current_content: list[
        openai_types.ChatCompletionContentPartTextParam
        | openai_types.ChatCompletionContentPartImageParam
        | openai_types.ChatCompletionContentPartInputAudioParam
    ] = []
    result: list[openai_types.ChatCompletionMessageParam] = []

    def flush_message_content() -> None:
        nonlocal current_content
        if current_content:
            if len(current_content) == 1 and current_content[0]["type"] == "text":
                result.append(
                    openai_types.ChatCompletionUserMessageParam(
                        role="user", content=current_content[0]["text"]
                    )
                )
            else:
                result.append(
                    openai_types.ChatCompletionUserMessageParam(
                        role="user", content=current_content
                    )
                )
            current_content = []

    for part in message.content:
        if part.type == "text":
            current_content.append(
                openai_types.ChatCompletionContentPartTextParam(
                    text=part.text, type="text"
                )
            )
        elif part.type == "image":
            url = (
                part.source.url
                if part.source.type == "url_image_source"
                else f"data:{part.source.mime_type};base64,{part.source.data}"
            )
            content = openai_types.ChatCompletionContentPartImageParam(
                type="image_url", image_url={"url": url, "detail": "auto"}
            )
            current_content.append(content)
        elif part.type == "audio":
            base_model_name = model_name(model_id, None)
            if base_model_name in MODELS_WITHOUT_AUDIO_SUPPORT:
                raise FeatureNotSupportedError(
                    feature="Audio inputs",
                    provider_id="openai",
                    message=f"Model '{model_id}' does not support audio inputs.",
                )

            if part.source.type == "base64_audio_source":
                audio_format = part.source.mime_type.split("/")[1]
                if audio_format not in ("wav", "mp3"):
                    raise FeatureNotSupportedError(
                        feature=f"Audio format: {audio_format}",
                        provider_id="openai",
                        message="OpenAI only supports 'wav' and 'mp3' audio formats.",
                    )  # pragma: no cover
                audio_content = openai_types.ChatCompletionContentPartInputAudioParam(
                    type="input_audio",
                    input_audio={
                        "data": part.source.data,
                        "format": audio_format,
                    },
                )
                current_content.append(audio_content)
        elif part.type == "tool_output":
            flush_message_content()
            result.append(
                openai_types.ChatCompletionToolMessageParam(
                    role="tool",
                    content=str(part.value),
                    tool_call_id=part.id,
                )
            )
        else:
            raise NotImplementedError(
                f"Unsupported user content part type: {part.type}"
            )
    flush_message_content()

    return result


def _encode_assistant_message(
    message: AssistantMessage, model_id: OpenAIModelId, encode_thoughts_as_text: bool
) -> openai_types.ChatCompletionAssistantMessageParam:
    """Convert Mirascope `AssistantMessage` to OpenAI `ChatCompletionAssistantMessageParam`."""

    if (
        message.provider_id in ("openai", "openai:completions")
        and message.provider_model_name
        == model_name(model_id=model_id, api_mode="completions")
        and message.raw_message
        and not encode_thoughts_as_text
    ):
        return cast(
            openai_types.ChatCompletionAssistantMessageParam, message.raw_message
        )

    text_params: list[openai_types.ChatCompletionContentPartTextParam] = []
    tool_call_params: list[openai_types.ChatCompletionMessageToolCallParam] = []
    for part in message.content:
        if part.type == "text":
            text_params.append(
                openai_types.ChatCompletionContentPartTextParam(
                    text=part.text, type="text"
                )
            )
        elif part.type == "tool_call":
            tool_call_params.append(
                openai_types.ChatCompletionMessageToolCallParam(
                    id=part.id,
                    type="function",
                    function={"name": part.name, "arguments": part.args},
                )
            )
        elif part.type == "thought":
            if encode_thoughts_as_text:
                text_params.append(
                    openai_types.ChatCompletionContentPartTextParam(
                        text="**Thinking:** " + part.thought, type="text"
                    )
                )
        else:
            raise NotImplementedError(f"Unsupported content type: {part.type}")

    content: str | list[openai_types.ChatCompletionContentPartTextParam] | None = None
    if len(text_params) == 1:
        content = text_params[0]["text"]
    elif text_params:
        content = text_params

    message_params: openai_types.ChatCompletionAssistantMessageParam = {
        "role": "assistant",
        "content": content,
    }
    if tool_call_params:
        message_params["tool_calls"] = tool_call_params

    return openai_types.ChatCompletionAssistantMessageParam(**message_params)


def _encode_message(
    message: Message, model_id: OpenAIModelId, encode_thoughts_as_text: bool
) -> list[openai_types.ChatCompletionMessageParam]:
    """Convert a Mirascope `Message` to OpenAI `ChatCompletionMessageParam` format.

    Args:
        message: A Mirascope message (system, user, or assistant)
        model_id: The model ID being used
        encode_thoughts: Whether to encode thoughts as text

    Returns:
        A list of OpenAI `ChatCompletionMessageParam` (may be multiple for tool outputs)
    """
    if message.role == "system":
        return [
            openai_types.ChatCompletionSystemMessageParam(
                role="system", content=message.content.text
            )
        ]
    elif message.role == "user":
        return _encode_user_message(message, model_id)
    elif message.role == "assistant":
        return [_encode_assistant_message(message, model_id, encode_thoughts_as_text)]
    else:
        raise ValueError(f"Unsupported role: {message.role}")  # pragma: no cover


@lru_cache(maxsize=128)
def _convert_tool_to_tool_param(
    tool: AnyToolSchema,
) -> openai_types.ChatCompletionToolParam:
    """Convert a single Mirascope `Tool` to OpenAI ChatCompletionToolParam with caching."""
    schema_dict = tool.parameters.model_dump(by_alias=True, exclude_none=True)
    schema_dict["type"] = "object"
    _base_utils.ensure_additional_properties_false(schema_dict)
    return openai_types.ChatCompletionToolParam(
        type="function",
        function={
            "name": tool.name,
            "description": tool.description,
            "parameters": schema_dict,
            "strict": tool.strict,
        },
    )


def _create_strict_response_format(
    format: Format[FormattableT],
) -> shared_openai_types.ResponseFormatJSONSchema:
    """Create OpenAI strict response format from a Mirascope Format.

    Args:
        format: The `Format` instance containing schema and metadata

    Returns:
        Dictionary containing OpenAI response_format specification
    """
    schema = format.schema.copy()

    _base_utils.ensure_additional_properties_false(schema)

    json_schema = JSONSchema(
        name=format.name,
        schema=schema,
        strict=True,
    )
    if format.description:
        json_schema["description"] = format.description

    return shared_openai_types.ResponseFormatJSONSchema(
        type="json_schema", json_schema=json_schema
    )


def encode_request(
    *,
    model_id: OpenAIModelId,
    messages: Sequence[Message],
    tools: Sequence[AnyToolSchema] | BaseToolkit[AnyToolSchema] | None,
    format: type[FormattableT] | Format[FormattableT] | None,
    params: Params,
) -> tuple[Sequence[Message], Format[FormattableT] | None, ChatCompletionCreateKwargs]:
    """Prepares a request for the `OpenAI.chat.completions.create` method."""
    if model_id.endswith(":responses"):
        raise FeatureNotSupportedError(
            feature="responses API",
            provider_id="openai:completions",
            model_id=model_id,
            message=f"Can't use completions client for responses model: {model_id}",
        )
    base_model_name = model_name(model_id, None)

    kwargs: ChatCompletionCreateKwargs = ChatCompletionCreateKwargs(
        {
            "model": base_model_name,
        }
    )
    encode_thoughts_as_text = False

    with _base_utils.ensure_all_params_accessed(
        params=params,
        provider_id="openai",
        unsupported_params=["top_k"],
    ) as param_accessor:
        if param_accessor.temperature is not None:
            kwargs["temperature"] = param_accessor.temperature
        if param_accessor.max_tokens is not None:
            kwargs["max_tokens"] = param_accessor.max_tokens
        if param_accessor.top_p is not None:
            kwargs["top_p"] = param_accessor.top_p

        if param_accessor.seed is not None:
            kwargs["seed"] = param_accessor.seed
        if param_accessor.stop_sequences is not None:
            kwargs["stop"] = param_accessor.stop_sequences
        if param_accessor.thinking is not None:
            thinking = param_accessor.thinking
            if thinking.get("encode_thoughts_as_text"):
                encode_thoughts_as_text = True

    tools = tools.tools if isinstance(tools, BaseToolkit) else tools or []

    openai_tools = [_convert_tool_to_tool_param(tool) for tool in tools]

    model_supports_strict = base_model_name not in MODELS_WITHOUT_JSON_SCHEMA_SUPPORT
    default_mode = "strict" if model_supports_strict else "tool"
    format = resolve_format(format, default_mode=default_mode)
    if format is not None:
        if format.mode == "strict":
            if not model_supports_strict:
                raise FormattingModeNotSupportedError(
                    formatting_mode="strict",
                    provider_id="openai",
                    model_id=model_id,
                )
            kwargs["response_format"] = _create_strict_response_format(format)
        elif format.mode == "tool":
            if tools:
                kwargs["tool_choice"] = "required"
            else:
                kwargs["tool_choice"] = {
                    "type": "function",
                    "function": {"name": FORMAT_TOOL_NAME},
                }
                kwargs["parallel_tool_calls"] = False
            format_tool_schema = format.create_tool_schema()
            openai_tools.append(_convert_tool_to_tool_param(format_tool_schema))
        elif (
            format.mode == "json"
            and base_model_name not in MODELS_WITHOUT_JSON_OBJECT_SUPPORT
        ):
            kwargs["response_format"] = {"type": "json_object"}

        if format.formatting_instructions:
            messages = _base_utils.add_system_instructions(
                messages, format.formatting_instructions
            )

    if openai_tools:
        kwargs["tools"] = openai_tools

    encoded_messages: list[openai_types.ChatCompletionMessageParam] = []
    for message in messages:
        encoded_messages.extend(
            _encode_message(message, model_id, encode_thoughts_as_text)
        )
    kwargs["messages"] = encoded_messages

    return messages, format, kwargs



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    ToolCall,
    ToolOutput,
    UserMessage,
)

sync_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 277,
                "output_tokens": 42,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "raw": "ResponseUsage(input_tokens=277, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=42, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=319)",
                "total_tokens": 319,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_vGDl14wuUgjKmbX3QvSPOUOl",
                            name="get_book_info",
                            args='{"isbn":"0-7653-1178-X"}',
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "arguments": '{"isbn":"0-7653-1178-X"}',
                            "call_id": "call_vGDl14wuUgjKmbX3QvSPOUOl",
                            "name": "get_book_info",
                            "type": "function_call",
                            "id": "fc_099b2074297131ee0068f966f805588195ab0a3266a3485eb2",
                            "status": "completed",
                        }
                    ],
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_vGDl14wuUgjKmbX3QvSPOUOl",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_099b2074297131ee0068f966f961e88195b52bc39cd2c63caf",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": """\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
""",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)
async_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 277,
                "output_tokens": 42,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "raw": "ResponseUsage(input_tokens=277, input_tokens_details=InputTokensDetails(cached_tokens=0), output_tokens=42, output_tokens_details=OutputTokensDetails(reasoning_tokens=0), total_tokens=319)",
                "total_tokens": 319,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_gvc1qWKn09rI7p2qrit1MWSl",
                            name="get_book_info",
                            args='{"isbn":"0-7653-1178-X"}',
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "arguments": '{"isbn":"0-7653-1178-X"}',
                            "call_id": "call_gvc1qWKn09rI7p2qrit1MWSl",
                            "name": "get_book_info",
                            "type": "function_call",
                            "id": "fc_001734ab3b50b4140068f9670a2160819488ed306c0f9ae85f",
                            "status": "completed",
                        }
                    ],
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_gvc1qWKn09rI7p2qrit1MWSl",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_001734ab3b50b4140068f9670b55e88194a0e018fd79284d50",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": """\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
""",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
        }
    }
)
stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_V1BAu6CPIaeWyMim4iiKLay2",
                            name="get_book_info",
                            args='{"isbn":"0-7653-1178-X"}',
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "arguments": '{"isbn":"0-7653-1178-X"}',
                            "call_id": "call_V1BAu6CPIaeWyMim4iiKLay2",
                            "name": "get_book_info",
                            "type": "function_call",
                            "id": "fc_0f1f1f1012662e2a0068f9671c3ef08194bfaebc9329d2baf8",
                            "status": "completed",
                        }
                    ],
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_V1BAu6CPIaeWyMim4iiKLay2",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_0f1f1f1012662e2a0068f9671dde588194a49f52a567767f75",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": """\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
""",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
            "usage": {
                "input_tokens": 277,
                "output_tokens": 42,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "raw": "None",
                "total_tokens": 319,
            },
            "n_chunks": 42,
        }
    }
)
async_stream_snapshot = snapshot(
    {
        "response": {
            "provider_id": "openai",
            "model_id": "openai/gpt-4o:responses",
            "provider_model_name": "gpt-4o:responses",
            "finish_reason": None,
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
"""
                    )
                ),
                UserMessage(
                    content=[
                        Text(
                            text="Please look up the book with ISBN 0-7653-1178-X and provide detailed info and a recommendation score"
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        ToolCall(
                            id="call_WeASSdshXbzOg7UDwf1RHvre",
                            name="get_book_info",
                            args='{"isbn":"0-7653-1178-X"}',
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "arguments": '{"isbn":"0-7653-1178-X"}',
                            "call_id": "call_WeASSdshXbzOg7UDwf1RHvre",
                            "name": "get_book_info",
                            "type": "function_call",
                            "id": "fc_0075d4e9136cabd30068f9672aa9b481938088b390244ed3ec",
                            "status": "completed",
                        }
                    ],
                ),
                UserMessage(
                    content=[
                        ToolOutput(
                            id="call_WeASSdshXbzOg7UDwf1RHvre",
                            name="get_book_info",
                            value="Title: Mistborn: The Final Empire, Author: Brandon Sanderson, Pages: 544, Published: 2006-07-25",
                        )
                    ]
                ),
                AssistantMessage(
                    content=[
                        Text(
                            text="""\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
"""
                        )
                    ],
                    provider_id="openai",
                    model_id="openai/gpt-4o:responses",
                    provider_model_name="gpt-4o:responses",
                    raw_message=[
                        {
                            "id": "msg_0075d4e9136cabd30068f9672c86288193a65c26fdd8affb48",
                            "content": [
                                {
                                    "annotations": [],
                                    "text": """\
{
  "title": "Mistborn: The Final Empire",
  "author": "Brandon Sanderson",
  "pages": 544,
  "publication_year": 2006
}\
""",
                                    "type": "output_text",
                                    "logprobs": [],
                                }
                            ],
                            "role": "assistant",
                            "status": "completed",
                            "type": "message",
                        }
                    ],
                ),
            ],
            "format": {
                "name": "BookSummary",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "pages": {"title": "Pages", "type": "integer"},
                        "publication_year": {
                            "title": "Publication Year",
                            "type": "integer",
                        },
                    },
                    "required": ["title", "author", "pages", "publication_year"],
                    "title": "BookSummary",
                    "type": "object",
                },
                "mode": "json",
                "formatting_instructions": """\
Respond only with valid JSON that matches this exact schema:
{
  "properties": {
    "title": {
      "title": "Title",
      "type": "string"
    },
    "author": {
      "title": "Author",
      "type": "string"
    },
    "pages": {
      "title": "Pages",
      "type": "integer"
    },
    "publication_year": {
      "title": "Publication Year",
      "type": "integer"
    }
  },
  "required": [
    "title",
    "author",
    "pages",
    "publication_year"
  ],
  "title": "BookSummary",
  "type": "object"
}\
""",
            },
            "tools": [
                {
                    "name": "get_book_info",
                    "description": "Look up book information by ISBN.",
                    "parameters": """\
{
  "properties": {
    "isbn": {
      "title": "Isbn",
      "type": "string"
    }
  },
  "required": [
    "isbn"
  ],
  "additionalProperties": false,
  "defs": null
}\
""",
                    "strict": False,
                }
            ],
            "usage": {
                "input_tokens": 277,
                "output_tokens": 42,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 0,
                "raw": "None",
                "total_tokens": 319,
            },
            "n_chunks": 42,
        }
    }
)



from mirascope import llm


@llm.call("openai/gpt-5")
def recommend_book(genre: str):
    return f"Please recommend a book in {genre}."


def main():
    response: llm.Response = recommend_book("fantasy")
    print(response.pretty())
    # [!code highlight:2]
    continuation: llm.Response = response.resume("Please explain your choice.")
    print(continuation.pretty())


main()



from .decode import (
    decode_async_stream,
    decode_response,
    decode_stream,
    model_name,
)
from .encode import encode_request

__all__ = [
    "decode_async_stream",
    "decode_response",
    "decode_stream",
    "encode_request",
    "model_name",
]



"""Base abstract interface for provider clients."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Mapping, Sequence
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TypeAlias, cast, overload
from typing_extensions import TypeVar, Unpack

from ...context import Context, DepsT
from ...exceptions import APIError, MirascopeLLMError
from ...formatting import Format, FormattableT
from ...messages import Message, UserContent, user
from ...responses import (
    AsyncChunkIterator,
    AsyncContextResponse,
    AsyncContextStreamResponse,
    AsyncResponse,
    AsyncStreamResponse,
    ChunkIterator,
    ContextResponse,
    ContextStreamResponse,
    Response,
    StreamResponse,
)
from ...tools import (
    AsyncContextTool,
    AsyncContextToolkit,
    AsyncTool,
    AsyncToolkit,
    ContextTool,
    ContextToolkit,
    Tool,
    Toolkit,
)
from .params import Params

if TYPE_CHECKING:
    from ...exceptions import MirascopeLLMError
    from ..provider_id import ProviderId

ProviderClientT = TypeVar("ProviderClientT")

Provider: TypeAlias = "BaseProvider[Any]"
"""Type alias for `BaseProvider` with any client type."""

ProviderErrorMap: TypeAlias = Mapping[
    type[Exception],
    "type[MirascopeLLMError] | Callable[[Exception], type[MirascopeLLMError]]",
]
"""Mapping from provider SDK exceptions to Mirascope error types.

Keys are provider SDK exception types (e.g., OpenAIError, AnthropicError).
Values can be:
- Error type: Simple 1:1 mapping (e.g., RateLimitError)
- Callable: Transform function returning error type based on exception details
"""


class BaseProvider(Generic[ProviderClientT], ABC):
    """Base abstract provider for LLM interactions.

    This class defines explicit methods for each type of call, eliminating
    the need for complex overloads in provider implementations.
    """

    id: ClassVar[ProviderId]
    """Provider identifier (e.g., "anthropic", "openai")."""

    default_scope: ClassVar[str | list[str]]
    """Default scope(s) for this provider when explicitly registered.

    Can be a single scope string or a list of scopes. For example:
    - "anthropic/" - Single scope
    - ["anthropic/", "openai/"] - Multiple scopes (e.g., for AWS Bedrock)
    """

    error_map: ClassVar[ProviderErrorMap]
    """Mapping from provider SDK exceptions to Mirascope error types.

    Values can be:
    - Error type: Simple 1:1 mapping (e.g., AnthropicRateLimitError -> RateLimitError)
    - Callable: Transform function returning error type based on exception details
                (e.g., lambda e: NotFoundError if e.code == "model_not_found" else BadRequestError)

    The mapping is walked via the exception's MRO, allowing both specific error handling
    and fallback to base SDK error types (e.g., AnthropicError -> APIError).
    """

    client: ProviderClientT

    @contextmanager
    def _wrap_errors(self) -> Generator[None, None, None]:
        """Wrap provider API calls and convert errors to Mirascope exceptions.

        Walks the exception's MRO to find the first matching error type in the
        provider's error_map, allowing both specific error handling and fallback
        to base SDK error types (e.g., AnthropicError -> APIError).
        """
        try:
            yield
        except Exception as e:
            # Walk MRO to find first matching error type in provider's error_map
            for error_class in type(e).__mro__:
                if error_class in self.error_map:
                    error_type_or_fn = self.error_map[error_class]

                    if isinstance(error_type_or_fn, type):
                        error_type = cast(type[MirascopeLLMError], error_type_or_fn)
                    else:
                        error_type = error_type_or_fn(e)

                    # Construct Mirascope error with metadata
                    error: MirascopeLLMError = error_type(str(e))
                    if isinstance(error, APIError):
                        error.status_code = self.get_error_status(e)
                    error.provider = self.id
                    error.original_exception = e
                    raise error from e

            # Not in error_map - not a provider error, re-raise as-is
            raise

    def _wrap_iterator_errors(self, iterator: ChunkIterator) -> ChunkIterator:
        """Wrap sync chunk iterator to handle errors during iteration."""
        # TODO: Consider moving this logic into BaseSyncStreamResponse if appropriate.
        with self._wrap_errors():
            yield from iterator

    async def _wrap_async_iterator_errors(
        self, iterator: AsyncChunkIterator
    ) -> AsyncChunkIterator:
        """Wrap async chunk iterator to handle errors during iteration."""
        # TODO: Consider moving this logic into BaseAsyncStreamResponse if appropriate.
        with self._wrap_errors():
            async for chunk in iterator:
                yield chunk

    @overload
    def call(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> Response:
        """Generate an `llm.Response` without a response format."""
        ...

    @overload
    def call(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> Response[FormattableT]:
        """Generate an `llm.Response` with a response format."""
        ...

    @overload
    def call(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Generate an `llm.Response` with an optional response format."""
        ...

    def call(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Generate an `llm.Response` by synchronously calling this client's LLM provider.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.Response` object containing the LLM-generated content.
        """
        with self._wrap_errors():
            return self._call(
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )

    @abstractmethod
    def _call(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Implementation for call(). Subclasses override this method."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None]:
        """Generate an `llm.ContextResponse` without a response format."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, FormattableT]:
        """Generate an `llm.ContextResponse` with a response format."""
        ...

    @overload
    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generate an `llm.ContextResponse` with an optional response format."""
        ...

    def context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generate an `llm.ContextResponse` by synchronously calling this client's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.ContextResponse` object containing the LLM-generated content.
        """
        with self._wrap_errors():
            return self._context_call(
                ctx=ctx,
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )

    @abstractmethod
    def _context_call(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Implementation for context_call(). Subclasses override this method."""
        ...

    @overload
    async def call_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse:
        """Generate an `llm.AsyncResponse` without a response format."""
        ...

    @overload
    async def call_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` with a response format."""
        ...

    @overload
    async def call_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` with an optional response format."""
        ...

    async def call_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate an `llm.AsyncResponse` by asynchronously calling this client's LLM provider.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.AsyncResponse` object containing the LLM-generated content.
        """
        with self._wrap_errors():
            return await self._call_async(
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )

    @abstractmethod
    async def _call_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Implementation for call_async(). Subclasses override this method."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None]:
        """Generate an `llm.AsyncContextResponse` without a response format."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, FormattableT]:
        """Generate an `llm.AsyncContextResponse` with a response format."""
        ...

    @overload
    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate an `llm.AsyncContextResponse` with an optional response format."""
        ...

    async def context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate an `llm.AsyncContextResponse` by asynchronously calling this client's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.AsyncContextResponse` object containing the LLM-generated content.
        """
        with self._wrap_errors():
            return await self._context_call_async(
                ctx=ctx,
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )

    @abstractmethod
    async def _context_call_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Implementation for context_call_async(). Subclasses override this method."""
        ...

    @overload
    def stream(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> StreamResponse:
        """Stream an `llm.StreamResponse` without a response format."""
        ...

    @overload
    def stream(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> StreamResponse[FormattableT]:
        """Stream an `llm.StreamResponse` with a response format."""
        ...

    @overload
    def stream(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Stream an `llm.StreamResponse` with an optional response format."""
        ...

    def stream(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generate an `llm.StreamResponse` by synchronously streaming from this client's LLM provider.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.StreamResponse` object for iterating over the LLM-generated content.
        """
        with self._wrap_errors():
            stream_response = self._stream(
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )
        stream_response._chunk_iterator = self._wrap_iterator_errors(  # pyright: ignore[reportPrivateUsage]
            stream_response._chunk_iterator  # pyright: ignore[reportPrivateUsage]
        )
        return stream_response

    @abstractmethod
    def _stream(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool] | Toolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Implementation for stream(). Subclasses override this method."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT, None]:
        """Stream an `llm.ContextStreamResponse` without a response format."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT, FormattableT]:
        """Stream an `llm.ContextStreamResponse` with a response format."""
        ...

    @overload
    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Stream an `llm.ContextStreamResponse` with an optional response format."""
        ...

    def context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate an `llm.ContextStreamResponse` by synchronously streaming from this client's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.ContextStreamResponse` object for iterating over the LLM-generated content.
        """
        with self._wrap_errors():
            stream_response = self._context_stream(
                ctx=ctx,
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )
        stream_response._chunk_iterator = self._wrap_iterator_errors(  # pyright: ignore[reportPrivateUsage]
            stream_response._chunk_iterator  # pyright: ignore[reportPrivateUsage]
        )
        return stream_response

    @abstractmethod
    def _context_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[Tool | ContextTool[DepsT]]
        | ContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Implementation for context_stream(). Subclasses override this method."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse:
        """Stream an `llm.AsyncStreamResponse` without a response format."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncStreamResponse[FormattableT]:
        """Stream an `llm.AsyncStreamResponse` with a response format."""
        ...

    @overload
    async def stream_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Stream an `llm.AsyncStreamResponse` with an optional response format."""
        ...

    async def stream_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generate an `llm.AsyncStreamResponse` by asynchronously streaming from this client's LLM provider.

        Args:
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.AsyncStreamResponse` object for asynchronously iterating over the LLM-generated content.
        """
        with self._wrap_errors():
            stream_response = await self._stream_async(
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )
        stream_response._chunk_iterator = self._wrap_async_iterator_errors(  # pyright: ignore[reportPrivateUsage]
            stream_response._chunk_iterator  # pyright: ignore[reportPrivateUsage]
        )
        return stream_response

    @abstractmethod
    async def _stream_async(
        self,
        *,
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool] | AsyncToolkit | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Implementation for stream_async(). Subclasses override this method."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: None = None,
        **params: Unpack[Params],
    ) -> AsyncContextStreamResponse[DepsT, None]:
        """Stream an `llm.AsyncContextStreamResponse` without a response format."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT],
        **params: Unpack[Params],
    ) -> AsyncContextStreamResponse[DepsT, FormattableT]:
        """Stream an `llm.AsyncContextStreamResponse` with a response format."""
        ...

    @overload
    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Stream an `llm.AsyncContextStreamResponse` with an optional response format."""
        ...

    async def context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate an `llm.AsyncContextStreamResponse` by asynchronously streaming from this client's LLM provider.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            messages: Messages to send to the LLM.
            tools: Optional tools that the model may invoke.
            format: Optional response format specifier.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            An `llm.AsyncContextStreamResponse` object for asynchronously iterating over the LLM-generated content.
        """
        with self._wrap_errors():
            stream_response = await self._context_stream_async(
                ctx=ctx,
                model_id=model_id,
                messages=messages,
                tools=tools,
                format=format,
                **params,
            )
        stream_response._chunk_iterator = self._wrap_async_iterator_errors(  # pyright: ignore[reportPrivateUsage]
            stream_response._chunk_iterator  # pyright: ignore[reportPrivateUsage]
        )
        return stream_response

    @abstractmethod
    async def _context_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        messages: Sequence[Message],
        tools: Sequence[AsyncTool | AsyncContextTool[DepsT]]
        | AsyncContextToolkit[DepsT]
        | None = None,
        format: type[FormattableT] | Format[FormattableT] | None = None,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Implementation for context_stream_async(). Subclasses override this method."""
        ...

    @overload
    def resume(
        self,
        *,
        model_id: str,
        response: Response,
        content: UserContent,
        **params: Unpack[Params],
    ) -> Response:
        """Resume an `llm.Response` without a response format."""
        ...

    @overload
    def resume(
        self,
        *,
        model_id: str,
        response: Response[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> Response[FormattableT]:
        """Resume an `llm.Response` with a response format."""
        ...

    @overload
    def resume(
        self,
        *,
        model_id: str,
        response: Response | Response[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Resume an `llm.Response` with an optional response format."""
        ...

    def resume(
        self,
        *,
        model_id: str,
        response: Response | Response[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> Response | Response[FormattableT]:
        """Generate a new `llm.Response` by extending another response's messages with additional user content.

        Args:
            model_id: Model identifier to use.
            response: Previous response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.Response` object containing the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.call()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return self.call(
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    async def resume_async(
        self,
        *,
        model_id: str,
        response: AsyncResponse,
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncResponse:
        """Resume an `llm.AsyncResponse` without a response format."""
        ...

    @overload
    async def resume_async(
        self,
        *,
        model_id: str,
        response: AsyncResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncResponse[FormattableT]:
        """Resume an `llm.AsyncResponse` with a response format."""
        ...

    @overload
    async def resume_async(
        self,
        *,
        model_id: str,
        response: AsyncResponse | AsyncResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Resume an `llm.AsyncResponse` with an optional response format."""
        ...

    async def resume_async(
        self,
        *,
        model_id: str,
        response: AsyncResponse | AsyncResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncResponse | AsyncResponse[FormattableT]:
        """Generate a new `llm.AsyncResponse` by extending another response's messages with additional user content.

        Args:
            model_id: Model identifier to use.
            response: Previous async response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.AsyncResponse` object containing the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.call_async()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return await self.call_async(
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextResponse[DepsT, None],
        content: UserContent,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None]:
        """Resume an `llm.ContextResponse` without a response format."""
        ...

    @overload
    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, FormattableT]:
        """Resume an `llm.ContextResponse` with a response format."""
        ...

    @overload
    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Resume an `llm.ContextResponse` with an optional response format."""
        ...

    def context_resume(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> ContextResponse[DepsT, None] | ContextResponse[DepsT, FormattableT]:
        """Generate a new `llm.ContextResponse` by extending another response's messages with additional user content.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            response: Previous context response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.ContextResponse` object containing the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.context_call()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return self.context_call(
            ctx=ctx,
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextResponse[DepsT, None],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None]:
        """Resume an `llm.AsyncContextResponse` without a response format."""
        ...

    @overload
    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, FormattableT]:
        """Resume an `llm.AsyncContextResponse` with a response format."""
        ...

    @overload
    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextResponse[DepsT, None]
        | AsyncContextResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Resume an `llm.AsyncContextResponse` with an optional response format."""
        ...

    async def context_resume_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextResponse[DepsT, None]
        | AsyncContextResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncContextResponse[DepsT, None] | AsyncContextResponse[DepsT, FormattableT]:
        """Generate a new `llm.AsyncContextResponse` by extending another response's messages with additional user content.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            response: Previous async context response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.AsyncContextResponse` object containing the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.context_call_async()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return await self.context_call_async(
            ctx=ctx,
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    def resume_stream(
        self,
        *,
        model_id: str,
        response: StreamResponse,
        content: UserContent,
        **params: Unpack[Params],
    ) -> StreamResponse:
        """Resume an `llm.StreamResponse` without a response format."""
        ...

    @overload
    def resume_stream(
        self,
        *,
        model_id: str,
        response: StreamResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> StreamResponse[FormattableT]:
        """Resume an `llm.StreamResponse` with a response format."""
        ...

    @overload
    def resume_stream(
        self,
        *,
        model_id: str,
        response: StreamResponse | StreamResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Resume an `llm.StreamResponse` with an optional response format."""
        ...

    def resume_stream(
        self,
        *,
        model_id: str,
        response: StreamResponse | StreamResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> StreamResponse | StreamResponse[FormattableT]:
        """Generate a new `llm.StreamResponse` by extending another response's messages with additional user content.

        Args:
            model_id: Model identifier to use.
            response: Previous stream response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.StreamResponse` object for streaming the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.stream()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return self.stream(
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    async def resume_stream_async(
        self,
        *,
        model_id: str,
        response: AsyncStreamResponse,
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse:
        """Resume an `llm.AsyncStreamResponse` without a response format."""
        ...

    @overload
    async def resume_stream_async(
        self,
        *,
        model_id: str,
        response: AsyncStreamResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse[FormattableT]:
        """Resume an `llm.AsyncStreamResponse` with a response format."""
        ...

    @overload
    async def resume_stream_async(
        self,
        *,
        model_id: str,
        response: AsyncStreamResponse | AsyncStreamResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Resume an `llm.AsyncStreamResponse` with an optional response format."""
        ...

    async def resume_stream_async(
        self,
        *,
        model_id: str,
        response: AsyncStreamResponse | AsyncStreamResponse[FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncStreamResponse | AsyncStreamResponse[FormattableT]:
        """Generate a new `llm.AsyncStreamResponse` by extending another response's messages with additional user content.

        Args:
            model_id: Model identifier to use.
            response: Previous async stream response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.AsyncStreamResponse` object for asynchronously streaming the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.stream_async()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return await self.stream_async(
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextStreamResponse[DepsT, None],
        content: UserContent,
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT, None]:
        """Resume an `llm.ContextStreamResponse` without a response format."""
        ...

    @overload
    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> ContextStreamResponse[DepsT, FormattableT]:
        """Resume an `llm.ContextStreamResponse` with a response format."""
        ...

    @overload
    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextStreamResponse[DepsT, None]
        | ContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Resume an `llm.ContextStreamResponse` with an optional response format."""
        ...

    def context_resume_stream(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: ContextStreamResponse[DepsT, None]
        | ContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> (
        ContextStreamResponse[DepsT, None] | ContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate a new `llm.ContextStreamResponse` by extending another response's messages with additional user content.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            response: Previous context stream response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.ContextStreamResponse` object for streaming the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.context_stream()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return self.context_stream(
            ctx=ctx,
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @overload
    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextStreamResponse[DepsT, None],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncContextStreamResponse[DepsT, None]:
        """Resume an `llm.AsyncContextStreamResponse` without a response format."""
        ...

    @overload
    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> AsyncContextStreamResponse[DepsT, FormattableT]:
        """Resume an `llm.AsyncContextStreamResponse` with a response format."""
        ...

    @overload
    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Resume an `llm.AsyncContextStreamResponse` with an optional response format."""
        ...

    async def context_resume_stream_async(
        self,
        *,
        ctx: Context[DepsT],
        model_id: str,
        response: AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT],
        content: UserContent,
        **params: Unpack[Params],
    ) -> (
        AsyncContextStreamResponse[DepsT, None]
        | AsyncContextStreamResponse[DepsT, FormattableT]
    ):
        """Generate a new `llm.AsyncContextStreamResponse` by extending another response's messages with additional user content.

        Args:
            ctx: Context object with dependencies for tools.
            model_id: Model identifier to use.
            response: Previous async context stream response to extend.
            content: Additional user content to append.
            **params: Additional parameters to configure output (e.g. temperature). See `llm.Params`.

        Returns:
            A new `llm.AsyncContextStreamResponse` object for asynchronously streaming the extended conversation.

        Note:
            Uses the previous response's tools and output format. This base method wraps
            around calling `client.context_stream_async()` with a messages array derived from the response
            messages. However, clients may override this with first-class resume logic.
        """
        messages = response.messages + [user(content)]
        return await self.context_stream_async(
            ctx=ctx,
            model_id=model_id,
            messages=messages,
            tools=response.toolkit,
            format=response.format,
            **params,
        )

    @abstractmethod
    def get_error_status(self, e: Exception) -> int | None:
        """Extract HTTP status code from provider-specific exception.

        Different SDKs store status codes differently (e.g., .status_code vs .code).
        Each provider implements this to handle their SDK's convention.

        Args:
            e: The exception to extract status code from.

        Returns:
            The HTTP status code if available, None otherwise.
        """
        ...



"""OpenTelemetry integration tests for `llm.Model.context_call`."""

from __future__ import annotations

from collections.abc import Generator

import pytest
from inline_snapshot import snapshot
from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
    InMemorySpanExporter,
)

from mirascope import llm, ops
from tests.ops.utils import span_snapshot


@pytest.fixture(autouse=True, scope="function")
def initialize() -> Generator[None, None, None]:
    """Initialize ops configuration and LLM instrumentation for each test."""
    ops.configure()
    ops.instrument_llm()
    yield
    ops.uninstrument_llm()


@pytest.mark.vcr()
def test_model_context_call_exports_genai_span(
    span_exporter: InMemorySpanExporter,
) -> None:
    """Ensure context_call emits a GenAI span with expected attributes."""
    model = llm.Model(model_id="openai/gpt-4o-mini")
    ctx = llm.Context(deps={"tenant": "kai"})
    messages = [
        llm.messages.system("You are a concise assistant."),
        llm.messages.user("Say hello to the user named Kai."),
    ]

    response = model.context_call(ctx=ctx, messages=messages)
    assert "Kai" in response.pretty()

    spans = span_exporter.get_finished_spans()
    assert len(spans) == 1
    span_dict = span_snapshot(spans[0])
    assert span_dict == snapshot(
        {
            "name": "chat openai/gpt-4o-mini",
            "kind": "CLIENT",
            "status": "UNSET",
            "attributes": {
                "gen_ai.operation.name": "chat",
                "gen_ai.provider.name": "openai",
                "gen_ai.request.model": "openai/gpt-4o-mini",
                "gen_ai.output.type": "text",
                "gen_ai.response.model": "openai/gpt-4o-mini",
                "gen_ai.response.finish_reasons": ["stop"],
                "gen_ai.response.id": "resp_0232fb1fc58d743a00691ae4934c6c8197a687889ce65332e6",
                "gen_ai.system_instructions": '[{"type":"text","content":"You are a concise assistant."}]',
                "gen_ai.input.messages": '[{"role":"user","parts":[{"type":"text","content":"Say hello to the user named Kai."}]}]',
                "gen_ai.output.messages": '[{"role":"assistant","parts":[{"type":"text","content":"Hello, Kai! How can I assist you today?"}],"finish_reason":"stop"}]',
            },
        }
    )



"""The `llm.call` decorator for turning `Prompt` functions into LLM calls."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Generic, cast, overload
from typing_extensions import Unpack

from ..context import DepsT
from ..formatting import Format, FormattableT
from ..models import Model
from ..prompts import (
    AsyncContextMessageTemplate,
    AsyncContextPrompt,
    AsyncMessageTemplate,
    AsyncPrompt,
    ContextMessageTemplate,
    ContextPrompt,
    MessageTemplate,
    Prompt,
    _utils,
)
from ..providers import ModelId, Params
from ..tools import (
    AsyncContextTool,
    AsyncContextToolkit,
    AsyncTool,
    AsyncToolkit,
    ContextTool,
    ContextToolkit,
    Tool,
    Toolkit,
    ToolT,
)
from ..types import P
from .calls import AsyncCall, AsyncContextCall, Call, ContextCall


@dataclass(kw_only=True)
class CallDecorator(Generic[ToolT, FormattableT]):
    """Decorator for converting a `MessageTemplate` into a `Call`.

    Takes a raw prompt function that returns message content and wraps it with tools,
    format, and a model to create a `Call` that can be invoked directly without needing
    to pass a model argument.

    The decorator automatically detects whether the function is async or context-aware
    and creates the appropriate `Call` variant (`Call`, `AsyncCall`, `ContextCall`, or `AsyncContextCall`).

    Conceptually: `CallDecorator` = `PromptDecorator` + `Model`
    Result: `Call` = `MessageTemplate` + tools + format + `Model`
    """

    model: Model
    """The default model to use with this call. May be overridden."""

    tools: Sequence[ToolT] | None
    """The tools that are included in the prompt, if any."""

    format: type[FormattableT] | Format[FormattableT] | None
    """The structured output format off the prompt, if any."""

    @overload
    def __call__(
        self: CallDecorator[AsyncTool | AsyncContextTool[DepsT], FormattableT],
        fn: AsyncContextMessageTemplate[P, DepsT],
    ) -> AsyncContextCall[P, DepsT, FormattableT]:
        """Decorate an async context prompt into an AsyncContextCall."""
        ...

    @overload
    def __call__(
        self: CallDecorator[Tool | ContextTool[DepsT], FormattableT],
        fn: ContextMessageTemplate[P, DepsT],
    ) -> ContextCall[P, DepsT, FormattableT]:
        """Decorate a context prompt into a ContextCall."""
        ...

    @overload
    def __call__(
        self: CallDecorator[AsyncTool, FormattableT], fn: AsyncMessageTemplate[P]
    ) -> AsyncCall[P, FormattableT]:
        """Decorate an async prompt into an AsyncCall."""
        ...

    @overload
    def __call__(
        self: CallDecorator[Tool, FormattableT], fn: MessageTemplate[P]
    ) -> Call[P, FormattableT]:
        """Decorate a prompt into a Call."""
        ...

    def __call__(
        self,
        fn: ContextMessageTemplate[P, DepsT]
        | AsyncContextMessageTemplate[P, DepsT]
        | MessageTemplate[P]
        | AsyncMessageTemplate[P],
    ) -> (
        ContextCall[P, DepsT, FormattableT]
        | AsyncContextCall[P, DepsT, FormattableT]
        | Call[P, FormattableT]
        | AsyncCall[P, FormattableT]
    ):
        """Decorates a prompt into a Call or ContextCall."""
        is_context = _utils.is_context_promptable(fn)
        is_async = _utils.is_async_promptable(fn)

        if is_context and is_async:
            tools = cast(
                Sequence[AsyncTool | AsyncContextTool[DepsT]] | None, self.tools
            )
            prompt = AsyncContextPrompt(
                fn=fn,
                toolkit=AsyncContextToolkit(tools=tools),
                format=self.format,
            )
            return AsyncContextCall(
                prompt=prompt,
                default_model=self.model,
            )
        elif is_context:
            tools = cast(Sequence[Tool | ContextTool[DepsT]] | None, self.tools)
            prompt = ContextPrompt(
                fn=fn,
                toolkit=ContextToolkit(tools=tools),
                format=self.format,
            )
            return ContextCall(
                prompt=prompt,
                default_model=self.model,
            )
        elif is_async:
            tools = cast(Sequence[AsyncTool] | None, self.tools)
            prompt = AsyncPrompt(
                fn=fn, toolkit=AsyncToolkit(tools=tools), format=self.format
            )
            return AsyncCall(
                prompt=prompt,
                default_model=self.model,
            )
        else:
            tools = cast(Sequence[Tool] | None, self.tools)
            prompt = Prompt(fn=fn, toolkit=Toolkit(tools=tools), format=self.format)
            return Call(
                prompt=prompt,
                default_model=self.model,
            )


@overload
def call(
    model: ModelId,
    *,
    tools: Sequence[ToolT] | None = None,
    format: type[FormattableT] | Format[FormattableT] | None = None,
    **params: Unpack[Params],
) -> CallDecorator[ToolT, FormattableT]:
    """Decorator for converting prompt functions into LLM calls.

    This overload accepts a model ID string and allows additional params.
    """
    ...


@overload
def call(
    model: Model,
    *,
    tools: Sequence[ToolT] | None = None,
    format: type[FormattableT] | Format[FormattableT] | None = None,
) -> CallDecorator[ToolT, FormattableT]:
    """Decorator for converting prompt functions into LLM calls.

    This overload accepts a Model instance and does not allow additional params.
    """
    ...


def call(
    model: ModelId | Model,
    *,
    tools: Sequence[ToolT] | None = None,
    format: type[FormattableT] | Format[FormattableT] | None = None,
    **params: Unpack[Params],
) -> CallDecorator[ToolT, FormattableT]:
    """Decorates a `MessageTemplate` to create a `Call` that can be invoked directly.

    The `llm.call` decorator is the most convenient way to use Mirascope. It transforms
    a raw prompt function (that returns message content) into a `Call` object that bundles
    the function with tools, format, and a model. The resulting `Call` can be invoked
    directly to generate LLM responses without needing to pass a model argument.

    The decorator automatically detects the function type:
    - If the first parameter is named `'ctx'` with type `llm.Context[T]` (or a subclass thereof),
      creates a `ContextCall`
    - If the function is async, creates an `AsyncCall` or `AsyncContextCall`
    - Otherwise, creates a regular `Call`

    The model specified in the decorator can be overridden at runtime using the
    `llm.model()` context manager. When overridden, the context model completely
    replaces the decorated model, including all parameters.

    Conceptual flow:
    - `MessageTemplate`: raw function returning content
    - `@llm.prompt`: `MessageTemplate` → `Prompt`
      Includes tools and format, if applicable. Can be called by providing a `Model`.
    - `@llm.call`: `MessageTemplate` → `Call`. Includes a model, tools, and format. The
      model may be created on the fly from a model identifier and optional params, or
      provided outright.

    Args:
        model: A model ID string (e.g., "openai/gpt-4") or a `Model` instance
        tools: Optional `Sequence` of tools to make available to the LLM
        format: Optional response format class (`BaseModel`) or Format instance
        **params: Additional call parameters (temperature, max_tokens, etc.)
            Only available when passing a model ID string

    Returns:
        A `CallDecorator` that converts prompt functions into `Call` variants
        (`Call`, `AsyncCall`, `ContextCall`, or `AsyncContextCall`)

    Example:

        Regular call:
        ```python
        from mirascope import llm

        @llm.call("openai/gpt-4")
        def recommend_book(genre: str):
            return f"Please recommend a book in {genre}."

        response: llm.Response = recommend_book("fantasy")
        print(response.pretty())
        ```

    Example:

        Context call:
        ```python
        from dataclasses import dataclass
        from mirascope import llm

        @dataclass
        class User:
            name: str
            age: int

        @llm.call("openai/gpt-4")
        def recommend_book(ctx: llm.Context[User], genre: str):
            return f"Recommend a {genre} book for {ctx.deps.name}, age {ctx.deps.age}."

        ctx = llm.Context(deps=User(name="Alice", age=15))
        response = recommend_book(ctx, "fantasy")
        print(response.pretty())
        ```
    """
    if isinstance(model, str):
        model = Model(model, **params)
    return CallDecorator(model=model, tools=tools, format=format)



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from .raw_client import AsyncRawOrganizationsClient, RawOrganizationsClient
from .types.organizations_create_response import OrganizationsCreateResponse
from .types.organizations_credits_response import OrganizationsCreditsResponse
from .types.organizations_get_response import OrganizationsGetResponse
from .types.organizations_list_response_item import OrganizationsListResponseItem
from .types.organizations_update_response import OrganizationsUpdateResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class OrganizationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawOrganizationsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawOrganizationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawOrganizationsClient
        """
        return self._raw_client

    def list(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[OrganizationsListResponseItem]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[OrganizationsListResponseItem]
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organizations.list()
        """
        _response = self._raw_client.list(request_options=request_options)
        return _response.data

    def create(
        self, *, name: str, slug: str, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationsCreateResponse:
        """
        Parameters
        ----------
        name : str
            a string at most 100 character(s) long

        slug : str
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsCreateResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organizations.create(
            name="name",
            slug="slug",
        )
        """
        _response = self._raw_client.create(name=name, slug=slug, request_options=request_options)
        return _response.data

    def get(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> OrganizationsGetResponse:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsGetResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organizations.get(
            id="id",
        )
        """
        _response = self._raw_client.get(id, request_options=request_options)
        return _response.data

    def update(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        slug: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OrganizationsUpdateResponse:
        """
        Parameters
        ----------
        id : str

        name : typing.Optional[str]
            a string at most 100 character(s) long

        slug : typing.Optional[str]
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsUpdateResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organizations.update(
            id="id",
        )
        """
        _response = self._raw_client.update(id, name=name, slug=slug, request_options=request_options)
        return _response.data

    def delete(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organizations.delete(
            id="id",
        )
        """
        _response = self._raw_client.delete(id, request_options=request_options)
        return _response.data

    def credits(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationsCreditsResponse:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsCreditsResponse
            Success

        Examples
        --------
        from mirascope.api._generated import Mirascope

        client = Mirascope()
        client.organizations.credits(
            id="id",
        )
        """
        _response = self._raw_client.credits(id, request_options=request_options)
        return _response.data


class AsyncOrganizationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawOrganizationsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawOrganizationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawOrganizationsClient
        """
        return self._raw_client

    async def list(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[OrganizationsListResponseItem]:
        """
        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[OrganizationsListResponseItem]
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organizations.list()


        asyncio.run(main())
        """
        _response = await self._raw_client.list(request_options=request_options)
        return _response.data

    async def create(
        self, *, name: str, slug: str, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationsCreateResponse:
        """
        Parameters
        ----------
        name : str
            a string at most 100 character(s) long

        slug : str
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsCreateResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organizations.create(
                name="name",
                slug="slug",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.create(name=name, slug=slug, request_options=request_options)
        return _response.data

    async def get(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationsGetResponse:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsGetResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organizations.get(
                id="id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.get(id, request_options=request_options)
        return _response.data

    async def update(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        slug: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> OrganizationsUpdateResponse:
        """
        Parameters
        ----------
        id : str

        name : typing.Optional[str]
            a string at most 100 character(s) long

        slug : typing.Optional[str]
            a string matching the pattern ^[a-z0-9][a-z0-9_-]*[a-z0-9]$

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsUpdateResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organizations.update(
                id="id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.update(id, name=name, slug=slug, request_options=request_options)
        return _response.data

    async def delete(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organizations.delete(
                id="id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.delete(id, request_options=request_options)
        return _response.data

    async def credits(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationsCreditsResponse:
        """
        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationsCreditsResponse
            Success

        Examples
        --------
        import asyncio

        from mirascope.api._generated import AsyncMirascope

        client = AsyncMirascope()


        async def main() -> None:
            await client.organizations.credits(
                id="id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.credits(id, request_options=request_options)
        return _response.data



from inline_snapshot import snapshot

from mirascope.llm import (
    AssistantMessage,
    SystemMessage,
    Text,
    UserMessage,
)

test_snapshot = snapshot(
    {
        "response": {
            "provider_id": "google",
            "model_id": "google/gemini-2.5-flash",
            "provider_model_name": "gemini-2.5-flash",
            "params": {},
            "finish_reason": None,
            "usage": {
                "input_tokens": 137,
                "output_tokens": 147,
                "cache_read_tokens": 0,
                "cache_write_tokens": 0,
                "reasoning_tokens": 110,
                "raw": """\
cache_tokens_details=None cached_content_token_count=None candidates_token_count=37 candidates_tokens_details=None prompt_token_count=137 prompt_tokens_details=[ModalityTokenCount(
  modality=<MediaModality.TEXT: 'TEXT'>,
  token_count=137
)] thoughts_token_count=110 tool_use_prompt_token_count=None tool_use_prompt_tokens_details=None total_token_count=284 traffic_type=None\
""",
                "total_tokens": 284,
            },
            "messages": [
                SystemMessage(
                    content=Text(
                        text="""\
Always recommend The Name of the Wind.
Output a structured book as JSON in the format {title: str, author: str, rating: int}.
The title should be in all caps, and the rating should always be the
lucky number 7.\
"""
                    )
                ),
                UserMessage(content=[Text(text="Please recommend a book to me!")]),
                AssistantMessage(
                    content=[
                        Text(
                            text='{"rating": 7, "title": "THE NAME OF THE WIND", "author": "Patrick Rothfuss"}'
                        )
                    ],
                    provider_id="google",
                    model_id="google/gemini-2.5-flash",
                    provider_model_name="gemini-2.5-flash",
                    raw_message={
                        "parts": [
                            {
                                "media_resolution": None,
                                "code_execution_result": None,
                                "executable_code": None,
                                "file_data": None,
                                "function_call": {
                                    "id": None,
                                    "args": {
                                        "rating": 7,
                                        "title": "THE NAME OF THE WIND",
                                        "author": "Patrick Rothfuss",
                                    },
                                    "name": "__mirascope_formatted_output_tool__",
                                    "partial_args": None,
                                    "will_continue": None,
                                },
                                "function_response": None,
                                "inline_data": None,
                                "text": None,
                                "thought": None,
                                "thought_signature": b'\n\xc5\x03\x01r\xc8\xda|\xc9cS\xb6\xb2\x81P\xba\xeb[\xa7\xb9B\x06-\x1a%\xe8\xa5\x84\xbd"y\x1e\xc3n\xfd\xde\x11\xac]P\xbf\xdcP"\x18\xe5\x14*%G\'\x03\xb1L\xe7[\x04\xdaF\xf1\xec\xc756d7\x92\xf9\xb7\x90!c8a\xa1r\xd0\xfb\xcbr\xff\nF\x80\xfa\xe6\xe1\xf3\xc0\x02\x19\xa4~hz\x0f\xeb\x00\x10\xdf\xf3\xbb\x10 j\xc6\x95Yop\xa1\xa98\xccwKSm(\x11iI\xe5\xd7\xe0\xcd\xa5\x7f\x92\xd9\xd84\xe3\x89D\xc5\xccx(9S2\x02\x8bA\xbf\x84\xc3C<\xbd\x9c\xff\xb5\xd5B_7\xc3U\x82h\xda\x98\xca\xec\xf96\xfdw\xe1D\xbb\n\x9b\xe2\xd3\x0e\xa2H\xdf\xf0\xd8\xa0\xab\xf5\xb1\xf6\x93\xe4\xcd\xea\xaf\xe5[\xe2\x91\x91<\x89#\x06\xc0\x00SNl\xc7S\xa8\xab{\xce\x83E\xac\xd5\xd39\x7f\xd3\xeb1\xfa[\xeaC\xb3\xb2\xc2%\xa4 \xef\x81\x0f\x80;\x90\x8b\x07\xc2v\x99\x9an\xcd\x8e,`\x8d8:\xdbl\x85\x1ec\x9f?\x0c\xf3\xa3\x18\x18\xb7%\xa2\x7f\xbd\x8e\x84n\xae\xee\x08\xb7\xf2\t_n\x08\xf4\xeb\xfa\xff\x01\x88\x85\xf2B\x10-c\xcf\xa0\x96\x88\xbb\x98I\xcd\xf0|\x17\x8eY\x18g\xccN-\xd7\xe3\xeb\xb6\xf8\x15\x81D\xff\xe1-\x14\xdd\xc1\x80BMa\x95\x1a\xe8\x1b/\xe3\x0fgJf\xe6C\xf2@\x88a7\x9e\xdf\x98\xf5\xe0\xc1g{4\x8c\x8a\xa6\xdb\xa5\xdah\x89\x12(\x16\\\x99\xa7\x05v\xb0+\xa8h\x95z.V\x02\xc5\'y\x0c\xcc\xd4\x88\xae\xd5v\xdd\xe4\x00\x96\xeb;}ER\xfb\x1a+\xdb\xdc\xb7\x1b\x9b\x96\x10\xa8>d\xeb\xbfW\xc8\x97u\x908\xd8\xcc\xfd]V9~\x99K[\x06\xcdq\xc7j\xfcY\xb6\xb1\x80?6-"P8\x93\xa3\x1e}\xf4\xfa\xd9\xc2\x8b',
                                "video_metadata": None,
                            }
                        ],
                        "role": "model",
                    },
                ),
            ],
            "format": {
                "name": "Book",
                "description": None,
                "schema": {
                    "properties": {
                        "title": {"title": "Title", "type": "string"},
                        "author": {"title": "Author", "type": "string"},
                        "rating": {"title": "Rating", "type": "integer"},
                    },
                    "required": ["title", "author", "rating"],
                    "title": "Book",
                    "type": "object",
                },
                "mode": "tool",
                "formatting_instructions": """\
Output a structured book as JSON in the format {title: str, author: str, rating: int}.
The title should be in all caps, and the rating should always be the
lucky number 7.\
""",
            },
            "tools": [],
        }
    }
)



from unittest.mock import MagicMock

import pytest

from mirascope.llm.messages import system
from mirascope.llm.providers.mlx.encoding.transformers import (
    _encode_message,  # pyright: ignore[reportPrivateUsage]
)


def test_encode_message_system() -> None:
    """Test encoding a system message."""
    msg = system("sys")
    encoded = _encode_message(msg)
    assert encoded == {"role": "system", "content": "sys"}


def test_encode_message_invalid() -> None:
    """Test encoding an invalid message role."""
    # Create a mock message with an unsupported role
    mock_msg = MagicMock()
    mock_msg.role = "invalid"
    with pytest.raises(ValueError, match="Unsupported message type"):
        _encode_message(mock_msg)
